package ui;

###############################################################################
#                                                                             #
# Slrnconf - a graphical configuration utility for SLRN                       #
# Copyright (C) 2000-2003 Torsten Schoenfeld                                  #
#                                                                             #
# This program is free software; you can redistribute it and/or modify it     #
# under the terms of the GNU General Public License as published by the Free  #
# Software Foundation; either version 2 of the License, or (at your option)   #
# any later version.                                                          #
#                                                                             #
# This program is distributed in the hope that it will be useful, but WITHOUT #
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or       #
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for   #
# more details.                                                               #
#                                                                             #
# You should have received a copy of the GNU General Public License along     #
# with this program; if not, write to the Free Software Foundation, Inc., 59  #
# Temple Place, Suite 330, Boston, MA 02111-1307 USA                          #
#                                                                             #
###############################################################################

use strict;
use Gtk qw(-init);

use preferences;

sub new {
  my ($class, $root, $help, $data) = @_;
  my $self = {};

  bless($self, $class);

  $self -> { root } = $root;
  $self -> { help } = $help;
  $self -> { data } = $data;

  $self -> { preferences } = preferences -> new($self);

  return $self;
}

sub bitmap_popup {
  my ($self, @variable) = @_;
  my $value = $self -> { preferences } -> get(@variable);

  my %bitmap = $self -> get_bitmap($variable[-1]);
  my %config = ();

  my $return = "NOK";

  #############################################################################

  foreach (sort { $b <=> $a } (keys(%bitmap))) {
    if ($value >= $_) {
      $value -= $_;
      $config{ $_ } = 1;
    }
    else {
      $config{ $_ } = 0;
    }
  }

  #############################################################################

  my $dialog = Gtk::Dialog -> new();

  my $button_ok = Gtk::Button -> new("Ok");
  my $button_cancel = Gtk::Button -> new("Cancel");

  #############################################################################

  $dialog -> signal_connect("delete_event", sub {
    Gtk -> main_quit();
    1;
  });

  $button_ok -> signal_connect("clicked", sub {
    $return = "OK";
    Gtk -> main_quit();
    1;
  });

  $button_cancel -> signal_connect("clicked", sub {
    Gtk -> main_quit();
    1;
  });

  #############################################################################

  foreach my $option(sort { $a <=> $b } (keys(%bitmap))) {
    my $button = Gtk::ToggleButton -> new(" $bitmap{ $option } ");

    $button -> set_active($config{ $option });

    $button -> signal_connect("toggled", sub {
      $config{ $option } = $button -> active();
      1;
    });

    $dialog -> vbox -> pack_start($button, 1, 1, 0);
  }

  #############################################################################

  $dialog -> action_area -> pack_start($button_ok, 1, 1, 0);
  $dialog -> action_area -> pack_start($button_cancel, 1, 1, 0);

  #############################################################################

  $dialog -> set_title($variable[-1]);
  $dialog -> set_border_width(7);
  $dialog -> set_modal(1);
  $dialog -> set_transient_for($self -> { main_window });
  $dialog -> show_all();

  Gtk -> main();

  #############################################################################

  $dialog -> destroy();

  if ($return eq "OK") {
    my $return_value = 0;

    foreach (keys(%config)) {
      $return_value += $_ if ($config{ $_ });
    }

    return $return_value;
  }

  return;
}

sub create_mainwindow {
  my ($self) = @_;

  $self -> { main_window } = Gtk::Window -> new();
  $self -> { main_box } = Gtk::VBox -> new(0, 0);
  $self -> { accel_group } = Gtk::AccelGroup -> new();

  $self -> { main_window } -> set_title("Slrnconf");

  $self -> { main_window } -> signal_connect("delete_event", sub {
    $self -> quit();
    1;
  });

  return 1;
}

sub create_menu {
  my ($self) = @_;

  my $item_factory = Gtk::ItemFactory -> new("Gtk::MenuBar", "<main>",
                                             $self -> { accel_group });

  $item_factory -> create_items(
    { path => "/_File",            type => "<Branch>" },
    { path => "/File/Open",    callback => sub { $self -> file_open(); }},
    { path => "/File/Save",    callback => sub { $self -> file_save(); }},
    { path => "/File/Save As", callback => sub { $self -> file_save_as(); }},
    { path => "/File/sep_1",       type => "<Separator>" },
    { path => "/File/Quit",    callback => sub { $self -> quit(); }},

    { path => "/_Help",            type => "<LastBranch>" },
    { path => "/Help/About",   callback => sub { $self -> show_file($self -> { root } . "/ABOUT"); }},
    { path => "/Help/License", callback => sub { $self -> show_file($self -> { root } . "/COPYING"); }},
    { path => "/Help/sep_2",       type => "<Separator>" },
    { path => "/Help/Help",    callback => sub { $self -> show_file($self -> { root } . "/README"); }}
  );

  $self -> { main_box } -> pack_start($item_factory -> get_widget("<main>"), 0,
                                      0, 0);

  return 1;
}

sub create_notebook {
  my ($self) = @_;

  $self -> { notebook } = Gtk::Notebook -> new();
  $self -> { notebook } -> set_tab_pos("top");

  $self -> { main_box } -> pack_start($self -> { notebook }, 1, 1, 0);

  return 1;
}

sub create_tabs {
  my ($self) = @_;

  my @tabs = (
    ["general",        "General"],
    ["group",          "Groups"],
    ["article",        "Articles"],
    ["post_and_mail",  "Post & Mail"],
    ["misc",           "Misc"],
    ["files_and_dirs", "Files & Dirs"],
    ["scoring",        "Scoring"],
    ["spool",          "Spool"],
    ["grouplens",      "Grouplens"],
    ["color",          "Color"],
    ["mono",           "Mono"]
  );

  foreach (@tabs) {
    my ($tab_id, $tab_label) = @$_;

    my @fields = $self -> get_tab($tab_id);

    my $container = Gtk::ScrolledWindow -> new(0, 0);
    my $table = Gtk::Table -> new($#fields, 3, 0);

    my $current_row = 0;

    if ($tab_id eq "color") {
      $table -> attach(Gtk::Label -> new("Foreground"), 1, 2, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);
      $table -> attach(Gtk::Label -> new("Background"), 2, 3, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);

      my $button = Gtk::Button -> new(" ? ");

      $button -> signal_connect("clicked", sub {
        $self -> show_file($self -> { help } . "/color");
        1;
      });

      $table -> attach($button, 3, 4, $current_row, $current_row + 1, "shrink",
                       "shrink", 5, 0);

      $current_row++;
    }

    elsif ($tab_id eq "mono") {
      $table -> attach(Gtk::Label -> new("1st"), 1, 2, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);
      $table -> attach(Gtk::Label -> new("2nd"), 2, 3, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);
      $table -> attach(Gtk::Label -> new("3rd"), 3, 4, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);
      $table -> attach(Gtk::Label -> new("4th"), 4, 5, $current_row,
                       $current_row + 1, "fill", "shrink", 0, 0);

      my $button = Gtk::Button -> new(" ? ");

      $button -> signal_connect("clicked", sub {
        $self -> show_file($self -> { help } . "/mono");
        1;
      });

      $table -> attach($button, 5, 6, $current_row, $current_row + 1, "shrink",
                       "shrink", 5, 0);

      $current_row++;
    }

    foreach (@fields) {
      my $caption = shift(@$_);

      #########################################################################

      if ($caption eq "<seperator>") {
        my $separator = Gtk::HSeparator -> new();

        $separator -> height(20);

        $table -> attach($separator, 0, 3, $current_row, $current_row + 1,
                         ["fill", "expand"], "shrink", 0, 0);

        $current_row++;

        next;
      }

      #########################################################################

      my @variable = split(/, /, shift(@$_));
      my $type = shift(@$_);

      my $label = Gtk::Label -> new($caption);

      $label -> set_alignment(1, 0);

      $table -> attach($label, 0, 1, $current_row, $current_row + 1, "fill",
                       "shrink", 5, 0);

      #########################################################################

      if ($type eq "text") {
        my $entry = Gtk::Entry -> new();

        my $signal_id = $entry -> signal_connect("changed", sub {
          $self -> { preferences } -> set(@variable, shift() -> get_text());
          1;
        });

        $self -> { preferences } ->
          attach_callback(@variable, sub {
            my ($value) = @_;

            $value = "" unless (defined($value));

            $entry -> signal_handler_block($signal_id);
            $entry -> set_text($value);
            $entry -> signal_handler_unblock($signal_id);
          });

        $table -> attach($entry, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "shrink", 0, 0);
      }

      #########################################################################

      elsif ($type eq "check") {
        my $check = Gtk::CheckButton -> new();

        my $signal_id = $check -> signal_connect("toggled", sub {
          $self -> { preferences } -> set(@variable,
                                          shift() -> active() ? 1 : 0);
          1;
        });

        $self -> { preferences } ->
          attach_callback(@variable, sub {
            my ($value) = @_;

            unless (defined($value)) {
              $value = 0;
            }
            elsif ($value != 0 && $value != 1) {
              $value = 1;
            }

            $check -> signal_handler_block($signal_id);
            $check -> set_active($value);
            $check -> signal_handler_unblock($signal_id);
          });

        $table -> attach($check, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "shrink", 0, 0);
      }

      #########################################################################

      elsif ($type eq "menu") {
        my $item_factory = Gtk::ItemFactory -> new("Gtk::OptionMenu", "<main>",
                                                   $self -> { accel_group });

        my @menu = $self -> get_menu($variable[-1]);

        $item_factory -> create_items(
          map {
            my ($label, $key) = @$_;

            $label =~ s/_/__/g;

            {
              path => ("/" . $label),
              callback => sub {
                $self -> { preferences } -> set(@variable, $key);
                1;
              }
            }
          } @menu
        );

        my $optionmenu = $item_factory -> get_widget("<main>");

        $optionmenu -> height(25);

        $self -> { preferences } -> attach_callback(@variable, sub {
                                      my ($value) = @_;

                                      if (defined($value)) {
                                        foreach (0 .. $#menu) {
                                          if ($menu[$_] -> [1] eq $value) {
                                            $optionmenu -> set_history($_);
                                          }
                                        }
                                      }
                                    });

        $table -> attach($optionmenu, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "fill", 0, 0);
      }

      #########################################################################

      elsif ($type eq "bitmap") {
        my $value = $self -> { preferences } -> get(@variable);
        $value = 0 unless (defined($value));

        my $button = Gtk::Button -> new($value);

        $button -> signal_connect("clicked", sub {
          my $new_value = $self -> bitmap_popup(@variable);

          if (defined($new_value)) {
            $self -> { preferences } -> set(@variable, $new_value);
            shift() -> child -> set($new_value);
          }

          1;
        });

        $self -> { preferences } -> attach_callback(@variable, sub {
                                      my ($value) = @_;

                                      if (defined($value)) {
                                        $button -> child -> set($value);
                                      }
                                    });

        $table -> attach($button, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "fill", 0, 0);
      }

      #########################################################################

      elsif ($type eq "list") {
        my $dialog_title = shift(@$_);
        my $dialog = shift(@$_);

        my $container = Gtk::HBox -> new(0, 0);

        #######################################################################

        my $clist_container = Gtk::ScrolledWindow -> new(0, 0);
        my $clist = Gtk::CList -> new(1);

        $self -> { preferences } -> attach_callback(@variable, sub {
                                      my ($value) = @_;

                                      if (defined($value)) {
                                        $clist -> clear();

                                        foreach (@{$value}) {
                                          $clist -> append($_);
                                        }
                                      }
                                    });

        $clist -> column_titles_hide();
        $clist -> set_selection_mode("single");
        $clist -> set_shadow_type("none");

        $clist -> signal_connect("button_press_event", sub {
          my ($clist, $event) = @_;

          if ($event -> { button } == 1 &&
              $event -> { type } eq "2button_press") {
            Gtk -> idle_add(sub {
              $self -> manage_list($clist, "edit", \@variable, $dialog_title,
                                   $dialog);
              0;
            });
          }

          1;
        });

        $clist -> signal_connect("button_release_event", sub {
          my ($clist, $event) = @_;

          if ($event -> { button } == 3 &&
              $event -> { type } eq "button_release") {
            $self -> popup($clist, \@variable, $dialog_title, $dialog);
          }

          1;
        });

        $clist_container -> add_with_viewport($clist);
        $clist_container -> set_policy("automatic", "automatic");

        #######################################################################

        my $button_box = Gtk::VBox -> new(1, 0);

        my $arrow_up = Gtk::Arrow -> new("up", "none");
        my $arrow_down = Gtk::Arrow -> new("down", "none");

        my $button_up = Gtk::Button -> new();
        my $button_down = Gtk::Button -> new();

        $button_up -> add($arrow_up);
        $button_down -> add($arrow_down);

        $button_up -> signal_connect("clicked", sub {
          $self -> manage_list($clist, "up", \@variable);
          1;
        }),

        $button_down -> signal_connect("clicked", sub {
          $self -> manage_list($clist, "down", \@variable);
          1;
        }),

        $button_box -> pack_start($button_up, 1, 1, 0);
        $button_box -> pack_start($button_down, 1, 1, 0);

        #######################################################################

        $container -> pack_start($clist_container, 1, 1, 0);
        $container -> pack_start($button_box, 0, 0, 0);

        $table -> attach($container, 1, 2, $current_row, $current_row + 1,
                         ["fill", "expand"], "shrink", 0, 0);
      }

      #########################################################################

      elsif ($type eq "color") {
        my $identifier = $variable[-1];

        foreach my $type (qw(fg bg)) {
          my @menu = $self -> get_colors();
          my $item_factory = Gtk::ItemFactory -> new("Gtk::OptionMenu",
                                                     "<main>",
                                                     $self -> { accel_group });

          $item_factory -> create_items(
            map {
              my $color = $_;

              {
                path => ("/" . $color),
                callback => sub {
                  $self -> { preferences } -> set("color",
                                                  $identifier . "_" . $type,
                                                  $color);
                  1;
                }
              }
            } @menu
          );

          my $optionmenu = $item_factory -> get_widget("<main>");

          $optionmenu -> height(25);

          $self -> { preferences } ->
            attach_callback("color", $identifier . "_" . $type, sub {
              my ($value) = @_;

              if (defined($value)) {
                foreach (0 .. $#menu) {
                  if ($menu[$_] eq $value) {
                    $optionmenu -> set_history($_);
                  }
                }
              }
            });

          if ($type eq "fg") {
            $table -> attach_defaults($optionmenu, 1, 2, $current_row,
                                      $current_row + 1);
          }
          elsif ($type eq "bg") {
            $table -> attach_defaults($optionmenu, 2, 3, $current_row,
                                      $current_row + 1);
          }
        }
      }

      #########################################################################

      elsif ($type eq "mono") {
        my $identifier = $variable[-1];

        foreach my $type (qw(one two three four)) {
          my @menu = $self -> get_monos();
          my $item_factory = Gtk::ItemFactory -> new("Gtk::OptionMenu",
                                                     "<main>",
                                                     $self -> { accel_group });

          $item_factory -> create_items(
            map {
              my $mono = $_;

              {
                path => ("/" . $mono),
                callback => sub {
                  $self -> { preferences } -> set("mono",
                                                  $identifier . "_" . $type,
                                                  $mono);
                  1;
                }
              }
            } @menu
          );

          my $optionmenu = $item_factory -> get_widget("<main>");

          $optionmenu -> height(25);

          $self -> { preferences } ->
            attach_callback("mono", $identifier . "_" . $type, sub {
              my ($value) = @_;

              if (defined($value)) {
                foreach (0 .. $#menu) {
                  if ($menu[$_] eq $value) {
                    $optionmenu -> set_history($_);
                  }
                }
              }
            });

          if ($type eq "one") {
            $table -> attach_defaults($optionmenu, 1, 2, $current_row,
                                      $current_row + 1);
          }
          elsif ($type eq "two") {
            $table -> attach_defaults($optionmenu, 2, 3, $current_row,
                                      $current_row + 1);
          }
          elsif ($type eq "three") {
            $table -> attach_defaults($optionmenu, 3, 4, $current_row,
                                      $current_row + 1);
          }
          elsif ($type eq "four") {
            $table -> attach_defaults($optionmenu, 4, 5, $current_row,
                                      $current_row + 1);
          }
        }
      }

      #########################################################################

      unless ($tab_id eq "color" || $tab_id eq "mono") {
        my $button = Gtk::Button -> new(" ? ");

        $button -> signal_connect("clicked", sub {
          $self -> show_file($self -> { help } . "/" . $variable[-1]);
          1;
        });

        $table -> attach($button, 2, 3, $current_row, $current_row + 1,
                         "shrink", "shrink", 5, 0);
      }

      $current_row++;
    }

    ###########################################################################

    $table -> border_width(5);

    $container -> set_usize(600, 250);
    $container -> set_policy("automatic", "automatic");
    $container -> border_width(1);

    $container -> add_with_viewport($table);

    $self -> { notebook } -> append_page($container,
                                         Gtk::Label -> new($tab_label));
  }

  return 1;
}

sub create_statusbar {
  my ($self) = @_;

  $self -> { statusbar } = Gtk::Statusbar -> new();
  $self -> { statusbar_id } = $self -> { statusbar } ->
                                get_context_id("statusbar");

  $self -> { main_box } -> pack_start($self -> { statusbar }, 0, 0, 0);

  $self -> rc_file();

  return 1;
}

sub dialog {
  my ($self, $dialog_title, $dialog, $old_input) = @_;

  my @old_input = split(/ => /, $old_input || "");
  my $input = [];
  my $return = "NOK";

  #############################################################################

  my @dialog_rows;

  foreach my $label(split(/, /, $dialog)) {
    my $dialog_row = { };

    if ($label =~ s/^\*//) {
      $dialog_row -> { browse } = 1;
    }

    $dialog_row -> { label } = $label;

    push(@dialog_rows, $dialog_row);
  }

  #############################################################################

  my $dialog_box = Gtk::Dialog -> new();

  my $table = Gtk::Table -> new($#dialog_rows, 3, 0);

  my $button_ok = Gtk::Button -> new("Ok");
  my $button_cancel = Gtk::Button -> new("Cancel");

  $dialog_box -> signal_connect("delete_event", sub {
    Gtk -> main_quit();
    1;
  });

  $button_ok -> signal_connect("clicked", sub {
    $return = "OK";
    Gtk -> main_quit();
    1;
  });

  $button_cancel -> signal_connect("clicked", sub {
    Gtk -> main_quit();
    1;
  });

  #############################################################################

  my $row = 0;

  foreach (@dialog_rows) {
    my $entry = Gtk::Entry -> new();
    my $label = Gtk::Label -> new($_ -> { label });

    my $local_row = $row;

    $entry -> signal_connect("changed", sub {
      $input -> [$local_row] = shift() -> get_text();
      1;
    });

    if (defined($old_input[$row])) {
      $entry -> set_text($old_input[$row]);
    }

    $label -> set_alignment(1, 0);

    $table -> attach_defaults($label, 0, 1, $row, $row + 1);
    $table -> attach_defaults($entry, 1, 2, $row, $row + 1);

    if ($_ -> { browse }) {
      my $button_browse = Gtk::Button -> new("...");

      $button_browse -> signal_connect("clicked", sub {
        $self -> file_dialog("open", $entry);
        1;
      });

      $table -> attach_defaults($button_browse, 2, 3, $row, $row + 1);
    }

    $row++;
  }

  #############################################################################

  $dialog_box -> action_area -> pack_start($button_ok, 1, 1, 0);
  $dialog_box -> action_area -> pack_start($button_cancel, 1, 1, 0);

  #############################################################################

  $table -> set_row_spacings(5);
  $table -> set_col_spacings(5);

  $dialog_box -> vbox -> pack_start($table, 1, 1, 0);

  $dialog_box -> set_title($dialog_title);
  $dialog_box -> set_border_width(7);
  $dialog_box -> set_modal(1);
  $dialog_box -> set_transient_for($self -> { main_window });
  $dialog_box -> show_all();

  Gtk -> main();

  #############################################################################

  $dialog_box -> destroy();

  return $return eq "OK" ? $input : [];
}

sub error {
  my ($self, $error) = @_;

  my $id = $self -> { statusbar } -> push($self -> { statusbar_id }, $error);

  Gtk -> timeout_add(10000, sub {
    $self -> { statusbar } -> remove($self -> { statusbar_id }, $id);
    0;
  });

  return 0;
}

sub file_dialog {
  my ($self, $action, $entry) = @_;

  #############################################################################

  my $file_selector = Gtk::FileSelection -> new(
    ($action eq "save") ?
      "Select a file for saving" :
      "Select the file to open");

  $file_selector -> signal_connect("delete_event", sub {
    Gtk -> main_quit();
    1;
  });

  $file_selector -> ok_button -> signal_connect("clicked", sub {
    my $file = $file_selector -> get_filename();

    if ($action eq "save" && -e $file) {
      return 1
        unless ($self ->
                  question("Overwrite file?",
                           "\nDo you really want to overwrite '$file'?\n"));
    }

    if (defined($entry) && ref($entry) eq "Gtk::Entry") {
      $entry -> set_text($file);
    }
    else {
      $self -> { _tmp_file } = $file;
    }

    Gtk -> main_quit();
    1;
  });

  $file_selector -> cancel_button -> signal_connect("clicked", sub {
    Gtk -> main_quit();
    1;
  });

  $file_selector -> set_modal(1);
  $file_selector -> set_transient_for($self -> { main_window });
  $file_selector -> show_all();

  #############################################################################

  Gtk -> main();

  $file_selector -> destroy();

  return delete($self -> { _tmp_file });
}

sub file_open {
  my ($self) = @_;

  my $file = $self -> file_dialog("open");

  if (defined($file) && -f $file) {
    $self -> rc_file($file);
    Gtk -> idle_add(sub {
      $self -> read_preferences();
      0;
    });
  }

  return 1;
}

sub file_save {
  my ($self) = @_;
  return $self -> { preferences } -> write();
}

sub file_save_as {
  my ($self) = @_;
  my $file = $self -> file_dialog("save");

  if (defined($file)) {

    $self -> rc_file($file);
    $self -> { preferences } -> write();
  }

  return 1;
}

sub finalize_mainwindow {
  my ($self) = @_;

  $self -> { main_window } -> add($self -> { main_box });
  $self -> { main_window } -> add_accel_group($self -> { accel_group });

  $self -> { main_window } -> show_all();

  return 1;
}

sub get_bitmap {
  my ($self, $bitmap) = @_;
  my $indicator = 0;

  my %bitmap = ();

  open(BITMAPS, $self -> { data } . "/ui_bitmaps") or
    die($self -> { data } . "/ui_bitmaps: $!\n");

  while (my $line = <BITMAPS>) {
    if ($line =~ m/^$bitmap$/) {
      $indicator = 1;

      next;
    }

    if ($indicator && $line =~ s/^  //) {
      chomp($line);

      my ($key, $value) = split(/\t+/, $line);

      $bitmap{ $key } = $value;
    }
    else {
      $indicator = 0;
    }
  }

  close(BITMAPS);

  return %bitmap;
}

sub get_colors {
  my ($self) = @_;

  local $/ = undef;

  open(COLORS, $self -> { data } . "/ui_colors") or
    die($self -> { data } . "/ui_colors: $!\n");

  my @colors = split(/\n/, <COLORS>);

  close(COLORS);

  return @colors;
}

sub get_menu {
  my ($self, $menu) = @_;
  my $indicator = 0;

  my ($line, @menu);

  open(MENUS, $self -> { data } . "/ui_menus") or
    die($self -> { data } . "/ui_menus: $!\n");

  while ($line = <MENUS>) {
    if ($line =~ m/^$menu$/) {
      $indicator = 1;
      next;
    }

    if ($indicator && $line =~ s/^  //) {
      chomp($line);
      push(@menu, [split(/\t+/, $line)]);
    }
    else {
      $indicator = 0;
    }
  }

  close(MENUS);

  return @menu;
}

sub get_monos {
  my ($self) = @_;

  local $/ = undef;

  open(MONOS, $self -> { data } . "/ui_monos") or
    die($self -> { data } . "/ui_monos: $!\n");

  my @monos = split(/\n/, <MONOS>);

  close(MONOS);

  return @monos;
}

sub get_tab {
  my ($self, $tab) = @_;
  my $indicator = 0;

  my ($line, @tab);

  open(TABS, $self -> { data } . "/ui_tabs") or
    die($self -> { data } . "/ui_tabs: $!\n");

  while ($line = <TABS>) {
    if ($line =~ m/^$tab$/) {
      $indicator = 1;
      next;
    }

    if ($indicator && $line =~ s/^  //) {
      chomp($line);
      push(@tab, [split(/\t+/, $line)]);
    }
    else {
      $indicator = 0;
    }
  }

  close(TABS);

  return @tab;
}

sub manage_list {
  my ($self, $clist, $action, $variable, $dialog_title, $dialog) = @_;

  #############################################################################

  if ($action eq "add") {
    my $input = $self -> dialog($dialog_title, $dialog);

    if (defined($input -> [0])) {
      $clist -> append(join(" => ", @{$input}));
    }
  }

  #############################################################################

  else {
    my $selection = $clist -> selection();

    if (defined($selection)) {
      if ($action eq "del") {
        $clist -> remove($selection);
      }

      elsif ($action eq "up") {
        $clist -> row_move($selection, $selection - 1);
      }

      elsif ($action eq "down") {
        $clist -> row_move($selection, $selection + 1);
      }

      elsif ($action eq "edit") {
        my $old_value = $clist -> get_text($selection, 0) || return 0;
        my $input = $self -> dialog($dialog_title, $dialog, $old_value);

        if (defined($input -> [0])) {
          $clist -> remove($selection);
          $clist -> insert($selection, join(" => ", @{$input}));
        }
      }
    }
  }

  #############################################################################

  my $i = 0;
  my $value;

  while (defined($value = $clist -> get_text($i, 0))) {
    $self -> { preferences } -> index_set(@{$variable}, $i++, $value);
  }

  #############################################################################

  return 1;
}

sub popup {
  my ($self, $clist, $variable, $dialog_title, $dialog) = @_;

  my $item_factory = Gtk::ItemFactory -> new("Gtk::Menu", "<main>",
                                             $self -> { accel_group });

  if (defined($clist -> selection())) {
    $item_factory -> create_items(
      { path => "/Add",    callback => sub { $self -> manage_list($clist, "add", $variable, $dialog_title, $dialog); }},
      { path => "/Delete", callback => sub { $self -> manage_list($clist, "del", $variable); }},
      { path => "/Edit",   callback => sub { $self -> manage_list($clist, "edit", $variable, $dialog_title, $dialog); }},
    );
  }
  else {
    $item_factory -> create_items(
      { path => "/Add", callback => sub { $self -> manage_list($clist, "add", $variable, $dialog_title, $dialog); }}
    );
  }

  $item_factory -> get_widget("<main>") -> popup(undef, undef, 0, 3, undef,
                                                 undef);

  return 1;
}

sub question {
  my ($self, $title, $question) = @_;

  my $dialog = Gtk::Dialog -> new();
  my $message = Gtk::Label -> new($question);

  my $yes_button = Gtk::Button -> new("Yes");
  my $no_button = Gtk::Button -> new("No");

  my $return = 0;

  $dialog -> signal_connect("delete_event", sub {
    Gtk -> main_quit();
    1;
  });

  $yes_button -> signal_connect("clicked", sub {
    $return = 1;
    Gtk -> main_quit();
    1;
  });

  $no_button -> signal_connect("clicked", sub {
    Gtk -> main_quit();
    1;
  });

  $dialog -> vbox -> pack_start($message, 1, 1, 0);
  $dialog -> action_area -> pack_start($yes_button, 1, 1, 0);
  $dialog -> action_area -> pack_start($no_button, 1, 1, 0);

  $dialog -> set_title($title);
  $dialog -> set_border_width(7);
  $dialog -> set_modal(1);
  $dialog -> set_transient_for($self -> { main_window });
  $dialog -> show_all();

  Gtk -> main();
  $dialog -> destroy();

  return $return;
}

sub quit {
  my ($self) = @_;

  if ($self -> { preferences } -> { _changed }) {
    return 1
      unless ($self -> question("Really quit?",
                                "\nReally quit and abandon all changes?\n"));
  }

  Gtk -> main_quit();

  return 1;
}

sub read_preferences {
  my ($self) = @_;

  my $window = Gtk::Window -> new();
  my $label = Gtk::Label -> new("Reading " . $self -> rc_file() . " ...");

  $window -> realize();

  $window -> window -> set_decorations(0);

  $window -> set_usize(250, 50);
  $window -> set_transient_for($self -> { main_window });

  $window -> add($label);
  $window -> show_all();

  $self -> { main_window } -> set_sensitive(0);

  Gtk -> timeout_add(100, sub {
    $self -> { preferences } -> read();
    $window -> destroy();
    $self -> { main_window } -> set_sensitive(1);
    0;
  });

  return 1;
}

sub rc_file {
  my ($self, $new_rc_file) = @_;

  if (defined($new_rc_file)) {
    $self -> { preferences } -> { rc_file } = $new_rc_file;
  }

  $self -> { main_window } ->
    set_title("Slrnconf - " . $self -> { preferences } -> { rc_file });

  return $self -> { preferences } -> { rc_file };
}

sub run {
  my ($self) = @_;

  $self -> create_mainwindow();
  $self -> create_menu();
  $self -> create_notebook();
  $self -> create_statusbar();
  $self -> create_tabs();

  $self -> finalize_mainwindow();

  Gtk -> idle_add(sub {
    $self -> read_preferences();
    0;
  });

  Gtk -> main();

  return 1;
}

sub show_file {
  my ($self, $file) = @_;

  my $window = Gtk::Window -> new();
  my $container = Gtk::HBox -> new(0, 0);
  my $text = Gtk::Text -> new();
  my $scrollbar = Gtk::VScrollbar -> new($text -> vadj);

  my $font = Gtk::Gdk::Font ->
    load("-misc-fixed-medium-r-normal-*-12-*-*-*-*-*-*-*");

  $window -> set_title("Slrnconf - $file");
  $window -> signal_connect("delete_event", sub {
    Gtk -> main_quit();
    1;
  });

  local $/ = undef;

  open(HELP, "<$file") or
    return $self -> error("Couldn't open '$file' for reading.");

  $text -> insert($font, undef, undef, <HELP>);

  close(HELP);

  $text -> set_editable(0);
  $text -> signal_connect("button_release_event", sub {
    my $button = $_[1] -> { button };
    my $value = $text -> vadj -> get_value();

    if ($button == 4) {
      $text -> vadj -> set_value($value - 90);
    }
    elsif ($button == 5) {
      $text -> vadj -> set_value($value + 90);
    }

    1;
  });

  $container -> set_usize(600, 200);
  $container -> pack_start($text, 1, 1, 0);
  $container -> pack_start($scrollbar, 0, 0, 0);

  $window -> add($container);
  $window -> set_modal(1);
  $window -> show_all();

  Gtk -> main();

  $window -> destroy();

  return 1;
}

1;
